/*
PrintIt Copyright (c) 2005-2006, Alexander R. Pruss
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (
INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (
INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/
#include <PalmOS.h>
#include "PrintIt.h"

#include "dither.c"

#define CONV_16_TO_5( x ) ( ( ( x ) * 4 + 8 ) / 15 )

#define MAX_DATE_AND_TIME  256
#define MAX_EXTRAS_IN_LINE  32 /* max extra data per scan line */
#define HEADER_SIZE 1024

static Char* trailer[ NUM_DRIVERS ] =
{
     "\x1B*rB\x0C",             /* PCL */
     "showpage\r\n\4"  /* PS */
};

static Char* resetString[ NUM_DRIVERS ] =
{
     "\x1B" "E", /* PCL */
     ""          /* PS */
};

static Char* scanLineHeader[ NUM_DRIVERS ] =
{
    "\x1B*b%uW", /* PCL */
    ""           /* PS */
};

static Char* scanLineTrailer[ NUM_DRIVERS ] =
{
     "", /* PCL */
     "\r\n" /* PS */
};


typedef UInt8 DitherEntry5[ 2 ];

DitherEntry5 dither5[ 5 ] = {
  { 0, 0 },
  { 1, 0 },
  { 1, 2 },
  { 3, 2 },
  { 3, 3 }
};

static TimeFormatType GetTimeFormat( void )
{
     UInt16 index;
     LmLocaleType loc;
     TimeFormatType tf;

     OmGetSystemLocale( &loc );

     if ( errNone != LmLocaleToIndex( &loc, &index ) ||
          errNone != LmGetLocaleSetting( index, lmChoiceTimeFormat, &tf, sizeof( TimeFormatType ) ) ) {
         return tfColonAMPM;
     }
     return tf;
}



static DateFormatType GetDateFormat( void )
{
     UInt16 index;
     LmLocaleType loc;
     DateFormatType df;

     OmGetSystemLocale( &loc );

     if ( errNone != LmLocaleToIndex( &loc, &index ) ||
          errNone != LmGetLocaleSetting( index, lmChoiceDateFormat, &df, sizeof( DateFormatType ) ) ) {
         return dfMDYLongWithComma;
     }
     return df;
}




UInt8* RenderLine4( DriverType driver, UInt8* out, UInt8* in, UInt16 ditherRow, Coord width, Boolean big, Boolean dataMatters )
{
    Coord x;

    switch ( driver ) {
        case DRIVER_PCL:
            if ( ! big ) {
                if ( ! dataMatters )
                    return out + width / 4;
                for ( x = 0 ; x < width ; x += 4, in += 2 ) {
                     *out++ = ( dither5[ CONV_16_TO_5( in[ 0 ] >> 4 ) ][ ditherRow ] << 6 ) |
                              ( dither5[ CONV_16_TO_5( in[ 0 ] & 0xF ) ][ ditherRow ] << 4 ) |
                              ( dither5[ CONV_16_TO_5( in[ 1 ] >> 4 ) ][ ditherRow ] << 2 ) |
                              ( dither5[ CONV_16_TO_5( in[ 1 ] & 0xF ) ][ ditherRow ] );
                }
            }
            else {
                if ( ! dataMatters )
                    return out + width / 2;
                for ( x = 0 ; x < width ; x += 2, in++ ) {
                     *out++ = ( dither16[ *in >> 4 ][ ditherRow ] << 4 ) |
                              ( dither16[ *in & 0xF ][ ditherRow ] );
                }
            }
            break;

        case DRIVER_PS:
            if ( ! dataMatters )
                return out + width;
            for ( x = 0 ; x < width ; x += 2, in++ ) {
                 UInt8 nibble;

                 nibble = 15 - ( in[ 0 ] >> 4 );

                 if ( nibble < 0xA )
                     *out++ = nibble + '0';
                 else
                     *out++ = nibble + ( 'A' - 10 );

                 nibble = 15 - ( in[ 0 ] & 0xF );

                 if ( nibble < 0xA )
                     *out++ = nibble + '0';
                 else
                     *out++ = nibble + ( 'A' - 10 );
            }
            break;
    }
    return out;
}



/* Get a pixel in range 0-4 */
UInt8 GetPixel( Coord x, Coord y, UInt16 maxValue )
{
    RGBColorType c;

    WinGetPixelRGB( x, y, &c );

    return maxValue - ( ( ( UInt16 )c.r + c.g + c.b ) * maxValue + 3 * 255 / 2 ) / ( 3 * 255 );
}



static void InsertDateTime( Char* s, UInt32 timeStamp, Boolean postScriptEscape )
{
    Char* dateAndTimeString;
    DateTimeType dt;

    TimSecondsToDateTime( timeStamp, &dt );

    dateAndTimeString = MemPtrNew( MAX_DATE_AND_TIME + 1 );

    if ( dateAndTimeString == NULL ) {
        StrCopy( s, "(out of memory)" );
        return;
    }

    TimeToAscii( dt.hour, dt.minute, GetTimeFormat(), dateAndTimeString );
    StrCat( dateAndTimeString, ", " );
    DateToAscii( dt.month, dt.day, dt.year, GetDateFormat(), 
        dateAndTimeString + StrLen( dateAndTimeString)  );

    if ( postScriptEscape ) {
        Char* p;

        p = dateAndTimeString;

        while ( *p ) {
            if ( ( UInt8 )*p < ' ' || 126 < ( UInt8 )*p ) {
                /* hope for the best encoding-wise */
                /* octal */
                *s++ = '\\';
                *s++ = ( ( ( UInt8 )*p ) >> 6 ) + '0';
                *s++ = ( ( ( ( UInt8 )*p ) >> 3 ) & 7 ) + '0';
                *s++ = ( ( ( UInt8 )*p ) & 7 ) + '0';
                p++;
                continue;
            }
            if ( *p == '(' || *p == ')' || *p == '\\' ) {
                *s++ = '\\';
            }
            *s++ = *p++;
        }
        *s = 0;
    }
    else {
        StrCopy( s, dateAndTimeString );
    }

    MemPtrFree( dateAndTimeString );
}



Boolean PrintScreenDriver( DriverType driver, void* data,
    ProtocolSendFunction SendFunction, UInt32 sendFunctionRef, Boolean dataMatters )
{
    Coord width;
    Coord height;
    Coord y;
    UInt16 rowBytes;
    UInt16 ditherMultiplier;
    UInt8* dataBuffer;
    UInt8* bitmapData;
    UInt16 byteWidth;
    UInt16 oldCoords;
    UInt16 bufferSize;
    RectangleType rect;
    WinHandle oldWin;
    Boolean   big;
    UInt16    sizeMult;
    Boolean   hiRes;
    Err       err;
    UInt32    version;
    WinHandle imageWindow;
    Boolean   result;
    PrintScreenData* inData;
    UInt16    scanLineHeaderLength;
    UInt16    pclResolution;

    inData = data;

    if ( inData->bitmap == NULL )
        return false;

    result = false;

    err = FtrGet( sysFtrCreator, sysFtrNumWinVersion, &version );
    hiRes = ( err == errNone && 4 <= version );

    oldWin    = WinSetDrawWindow( WinGetDisplayWindow() );
    oldCoords = hiRes ? WinSetCoordinateSystem( kCoordinatesNative ) : 72;

    imageWindow = NULL;
    dataBuffer  = NULL;

    BmpGetDimensions( inData->bitmap, &width, &height, &rowBytes );

    if (  width % 4 != 0 || BmpGetBitDepth( inData->bitmap ) != 4 )
        goto CLOSE;

    bitmapData = BmpGetBits( inData->bitmap );

    if ( width <= 160 )
        sizeMult = 2;
    else
        sizeMult = 1;

    big = inData->big;

    // assume width is even
    
    pclResolution = 300;

    switch ( driver ) {
        case DRIVER_PCL:
            if ( big && inData->force600DPI ) {
                big           = false;
                pclResolution = 600;
            }

            if ( big )
                ditherMultiplier = 4;
            else
                ditherMultiplier = 2;

            byteWidth = width * ditherMultiplier / 8;
            break;

        case DRIVER_PS:
            ditherMultiplier = 1;
            byteWidth        = width;
            break;

        default:
            goto CLOSE;
    }

    bufferSize = byteWidth + MAX_EXTRAS_IN_LINE;

    if ( bufferSize < HEADER_SIZE )
        bufferSize = HEADER_SIZE;

    dataBuffer = MemPtrNew( bufferSize );

    if ( dataBuffer == NULL ) {
        goto CLOSE;
    }

    switch ( driver ) {
        case DRIVER_PCL:
            StrCopy( dataBuffer, "\x1B" "E" "\x1B(4101T\x1B(s1B" "\x1B&a720V\x1B&a720H" ); // Switch to Bold CG-Times, (72pt,72pt)
            InsertDateTime( dataBuffer + StrLen( dataBuffer ), inData->timeStamp, false );
            StrPrintF( dataBuffer + StrLen( dataBuffer ), "\x1B&a960V\x1B&a720H" "\x1B*t%uR\x1B*r1A\x1B*b0M", pclResolution / sizeMult );
            break;
        case DRIVER_PS:
            StrPrintF( dataBuffer, "%%!PS-Adobe-1.0\r\n"
                                   "%%%% If you see this text, you're using the PostScript driver\r\n"
                                 "%%%% with a non-Postscript printer.  Oops!\r\n"
                                 "/w %u def "
                                 "/h %u def "
                                 "/r %u def "
                                 "/Times-Bold findfont "
                                 "12 scalefont "
                                 "setfont "
                                 "72 732 moveto\r\n"
                                 "(", width, height, big ? ( 75 / sizeMult ) : ( 150 / sizeMult ) );

            InsertDateTime( dataBuffer + StrLen( dataBuffer ), inData->timeStamp, true );

            StrPrintF( dataBuffer + StrLen( dataBuffer ), ") show\r\n"
                                 "72 720 h r div 72 mul sub translate "
                                 "w r div 72 mul " "h r div 72 mul scale "
                                 "w h 4 [w 0 0 h neg 0 h]\r\n"
                                 "{currentfile %u string readhexstring pop} image\r\n",
                                 byteWidth / 2 );
            break;
    }

    SendFunction( sendFunctionRef, dataBuffer, -1 );

    if ( ! scanLineHeader[ driver ][ 0 ] ) {
        *dataBuffer = 0;
    }
    else {
        StrPrintF( dataBuffer, scanLineHeader[ driver ], byteWidth );
    }
    scanLineHeaderLength = StrLen( dataBuffer );

    for ( y = 0 ; y < height ; y++ ) {
        UInt8* out;
        UInt16 ditherRow;

        for ( ditherRow = 0 ; ditherRow < ditherMultiplier ; ditherRow++ ) {
            UInt8* in;

            in  = bitmapData + ( UInt32 )rowBytes * y;

            out = dataBuffer + scanLineHeaderLength;

            out = RenderLine4( driver, out, in, ditherRow, width, big, dataMatters );

            StrCopy( out, scanLineTrailer[ driver ] );
            out += StrLen( out );

            if ( ! SendFunction( sendFunctionRef, dataBuffer, out - dataBuffer ) )
                break;
        }

        if ( ditherRow < ditherMultiplier )
            break;

        if ( dataMatters ) {
            rect.topLeft.x = 0;
            rect.topLeft.y = y;
            rect.extent.x  = width;
            rect.extent.y  = 1;

            WinInvertRectangle( &rect, 0 );
        }
    }

    if ( y == height ) {
        if ( dataMatters ) {
            /* This compensates for some PSUS4 issue with
               LPD/LPR that I don't understand */
            SysTaskDelay( SysTicksPerSecond() / 10 );
        }
        result = SendFunction( sendFunctionRef, trailer[ driver ], -1 );
    }

    if ( dataMatters ) {
        rect.topLeft.x = 0;
        rect.topLeft.y = 0;
        rect.extent.x  = width;
        rect.extent.y  = y;
        WinInvertRectangle( &rect, 0 );
    }

    if ( result ) {
        result = SendFunction( sendFunctionRef, resetString[ driver ], -1 );
    }

CLOSE:
    if ( dataBuffer != NULL )
        MemPtrFree( dataBuffer );
    if ( hiRes )
        WinSetCoordinateSystem( oldCoords );
    if ( oldWin != NULL )
        WinSetDrawWindow( oldWin );
    if ( imageWindow != NULL )
        WinDeleteWindow( imageWindow, false );

    return result;
}


